#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "genlib.h"
#include "KernelDensity.h"

#ifndef M_PI
	#define M_PI 3.14159265358979323846
#endif

double		DefBandWidth(KERNELD* KD)
{
	double Ret, SumX, SumXX;
	int Index;
	
	SumX = 0;
	SumXX = 0;

	for(Index=0;Index<KD->NoData;Index++)
	{
		SumX += KD->Data[Index];
		SumXX += KD->Data[Index] * KD->Data[Index];
	}

	SumX = SumX / KD->NoData;
	SumXX = SumXX / KD->NoData;

	Ret = sqrt(SumXX - (SumX * SumX));

	Ret = Ret*(pow((3.0*KD->NoData/4.0),-0.2));

	return Ret;
}

double		CaclIQR(int Size, double *List)
{
	int Pos;
	double U, L;

	Pos = (int)(Size * 0.25);
	L = List[Pos];

	Pos = (int)(Size * 0.75);
	U = List[Pos];

	return U - L;
}

static int DComp (const void * a, const void * b)
{
	double Da, Db;

	Da = *(double*)a;
	Db = *(double*)b;

	if(Da == Db)
		return 0;

	if(Da > Db)
		return 1;

	return -1;
}

double		BandWidthSilverman(KERNELD *KD)
{
	double M, SD, IQR;
	double Ret;

	IQR = CaclIQR(KD->NoData, KD->Data) / 1.349;
	SD = KD->SD;

	M = SD;
	if(IQR < SD && IQR != 0)
		M = IQR;
	
	Ret = 0.9 * M * pow(KD->NoData, -0.2);

	return Ret;
}

double	BandWidthScott(KERNELD *KD)
{
	double M, SD, IQR;
	double Ret;

	IQR = CaclIQR(KD->NoData, KD->Data) / 1.349;
	SD = KD->SD;

	M = SD;
	if(IQR < SD && IQR != 0)
		M = IQR;
	
	Ret = 1.059 * M * pow(KD->NoData, -0.2);

	return Ret;
}

void		SetBandWidthPost(KERNELD *KD)
{
	KD->Min = KD->Data[0];
	KD->Max = KD->Data[KD->NoData-1];
	
	KD->Max = KD->Max + (DEF_EXT * KD->BandWidth);
	KD->Min = KD->Min - (DEF_EXT * KD->BandWidth);
}


void		SetBandWidth(double BandWidth, KERNELD *KD)
{
	KD->BandWidth = BandWidth;

	SetBandWidthPost(KD);
}

void		SetBandWidthSilverman(KERNELD *KD)
{
	KD->BandWidth = BandWidthSilverman(KD);
	SetBandWidthPost(KD);
}

void		SetBandWidthScott(KERNELD *KD)
{
	KD->BandWidth = BandWidthScott(KD);
	SetBandWidthPost(KD);
}


void		SetBandWidthDef(KERNELD *KD)
{
	KD->BandWidth = DefBandWidth(KD);
	SetBandWidthPost(KD);
}

void		CaclStats(KERNELD *KD)
{
	double Sum;
	int Index, No;

	No = KD->NoData;

	Sum = 0;
	for(Index=0;Index<KD->NoData;Index++)
		Sum += KD->Data[Index];
	KD->Mean = Sum / No;

	Sum = 0;
	for(Index=0;Index<KD->NoData;Index++)
		Sum += (KD->Data[Index] - KD->Mean) * (KD->Data[Index] - KD->Mean);
	Sum = Sum * (1.0 / (No-1));
	KD->SD = sqrt(Sum);
	
	if(KD->NoData % 2 == 0)
		KD->Median = (KD->Data[No/2] + KD->Data[No/2 - 1]) / 2.0;
	else
		KD->Median = KD->Data[No/2];
}

KERNELD*	CreateKD(int NoData, double *Data)
{
	KERNELD* Ret;
//	int Index;

	Ret = (KERNELD*)malloc(sizeof(KERNELD));
	if(Ret == NULL)
		MallocErr();

	Ret->NoData = NoData;
	Ret->Data = (double*)malloc(sizeof(double) * NoData);
	if(Ret->Data == NULL)
		MallocErr();
	memcpy(Ret->Data, Data, sizeof(double) * NoData);

	qsort(Ret->Data, NoData, sizeof(double), DComp);

	CaclStats(Ret);

	SetBandWidthDef(Ret);

	return Ret;
}

void		FreeKD(KERNELD* KD)
{
	free(KD->Data);
	free(KD);
}


double GaussPDF(double x, double m, double s)
{
	
	double z = (x - m)/s;
	return exp(-0.5*z*z)/(s*sqrt( 2.0*M_PI));

}

double		KDPDF(double X, KERNELD *KD)
{
	double Ret;
	int Index;

	Ret = 0;
	for(Index=0;Index<KD->NoData;Index++)
		Ret += GaussPDF(X, KD->Data[Index], KD->BandWidth);

	Ret = Ret / KD->NoData;

	return Ret;
}

double GaussCDF(double x, double m, double s)
{
	// Abramowitz Stegun Normal Cumulative Distribution Function	
	//double z = abs(x - m)/s;
	double z = fabs(x - m)/s;
  	double t = 1.0/(1.0 + 0.2316419*z);
  	double y = t*( 0.319381530 + t*(-0.356563782 + t*(1.781477937 + t*(-1.821255978 + t*1.330274429 ))));

	if(x >= m)
		return 1.0 - GaussPDF(x,m,s)*y*s;

	return GaussPDF(x,m,s)*y*s;
}

double		KDCDF(double X, KERNELD *KD)
{
	double Ret;
	int Index;

	Ret = 0;
	for(Index=0;Index<KD->NoData;Index++)
		Ret += GaussCDF(X, KD->Data[Index], KD->BandWidth);

	Ret = Ret / KD->NoData;

	return Ret;
}



double		FindMode(KERNELD *KD, int Steps)
{
	double MX, MPdf;
	double X, Pdf, StepSize;
	
	StepSize = (KD->Max - KD->Min) / Steps;

	MX = KD->Min;
	MPdf = KDPDF(MX, KD);

	for(X=KD->Min;X<KD->Max;X+=StepSize)
	{
		Pdf = KDPDF(X, KD);
		if(Pdf > MPdf)
		{
			MPdf = Pdf;
			MX = X;
		}

//		printf("%f\t%12.42f\n", X, Pdf);
	}
//	printf("\n\n\n");
	return MX;
}
